(use yasos testeez)
 
 ;;----------------------------
 ;; general operations
 ;;----------------------------
 
 (define-operation (print-obj obj port)
   (format port
     ;; if an instance does not have a print-obj operation..
     (if (instance? obj) "#<INSTANCE>~%" "#<NOT-AN-INSTANCE: ~s>~%") obj))
 
 (define-operation (size-obj obj)
   ;; default behavior
   (cond
     ((vector? obj) (vector-length obj))
     ((list? obj) (length obj))
     ((pair? obj) 2)
     ((string? obj) (string-length obj))
     ((char? obj) 1)
     (else
       (error "Operation not supported: size-obj" obj))))
 
 ;;----------------------
 ;; point interface
 ;;----------------------
 
 (define-predicate point?) ;; answers #f  by default
 (define-operation (x obj))
 (define-operation (y obj))
 (define-operation (set-x! obj new-x))
 (define-operation (set-y! obj new-y))
 
 ;;--------------------------------
 ;; point implementation
 ;;--------------------------------
 
 (define (make-point the-x the-y)
   (object
     ((point? self) #t) ;; yes, this is a point object
     ((x self) the-x)
     ((y self) the-y)
     ((set-x! self val)
       (set! the-x val)
       the-x)
     ((set-y! self val)
       (set! the-y val)
       the-y)
     ((size-obj self) 2)
     ((print-obj self port)
       (format port "#<point: ~a ~a>~%" (x self) (y self)))))
 
 ;;-----------------------------------------
 ;; 3D point interface additions
 ;;-----------------------------------------
 
 (define-predicate point-3d?) ;; #f by defualt
 (define-operation (z obj))
 (define-operation (set-z! obj new-z))
 
 ;;------------------------------------
 ;; 3D point implementation
 ;;------------------------------------
 
 (define (make-point-3d the-x the-y the-z)
   (object-with-ancestors ( (a-point (make-point the-x the-y)) )
     ((point-3d? self) #t)
     ((z self) the-z)
     ((set-z! self val) (set! the-z val) the-z)
     ;; override inherited size-obj and print-obj operations
     ((size-obj self) 3)
     ((print-obj self port)
       (format port "#<3d-point: ~a ~a ~a>~%" (x self) (y self) (z self)))))
 
 ;;;-----------------------
 ;; person interface
 ;;------------------------
 
 (define-predicate person?)
 (define-operation (name obj))
 (define-operation (age obj))
 (define-operation (set-age! obj new-age))
 (define-operation (ssn obj password)) ;; Social Security # is protected
 (define-operation (new-password obj old-passwd new-passwd))
 (define-operation (bad-password obj bogus-passwd)
   ;; assume internal (design) error
   (error (format #f "Bad Password: ~s given to ~a~%"
           bogus-passwd
           (print-obj obj #f))))
 
 ;;----------------------------------
 ;; person implementation
 ;;----------------------------------
 
 (define (make-person a-name an-age a-ssn the-password)
   (object
     ((person? self) #t)
     ((name self) a-name)
     ((age self) an-age)
     ((set-age! self val) (set! an-age val) an-age)
     ((ssn self password)
       (if (equal? password the-password)
         a-ssn
         (bad-password self password)))
     ((new-password self old-passwd new-passwd)
       (cond
         ((equal? old-passwd the-password) (set! the-password new-passwd) self)
         (else (bad-password self old-passwd))))
     ((bad-password self bogus-passwd)
       (format #t "Bad password: ~s~%" bogus-passwd)) ;; let user recover
     ((print-obj self port)
       (format port "#<Person: ~a age: ~a>~%" (name self) (age self)))))
 
 ;;;---------------------------------------------------------------
 ;; account-history and bank-account interfaces
 ;;----------------------------------------------------------------
  
 (define-predicate bank-account?)
 (define-operation (current-balance obj pin))
 (define-operation (add obj amount))
 (define-operation (withdraw obj amount pin))
 (define-operation (get-pin obj master-password))
 (define-operation (get-account-history obj master-password))
 
 ;;----------------------------------------------
 ;; account-history implementation
 ;;----------------------------------------------
 
 ;; put access to bank database and report generation here
 (define (make-account-history initial-balance a-pin master-password)
   ;; history is a simple list of balances -- no transaction times
   (letrec 
     ((history (list initial-balance))
      (balance (lambda () (car history))) ; balance is a function
      (remember
        (lambda (datum) (set! history (cons datum history)))))
     (object
       ((bank-account? self) #t)
       ((add self amount) ;; bank will accept money without a password
         (remember (+ amount (balance)))
         ;; print new balance
         (format #t "New balance: $~a~%" (balance)))
       ((withdraw self amount pin)
         (cond
           ((not (equal? pin a-pin)) (bad-password self pin))
           ((< (- (balance) amount) 0)
             (format 
               #t
               "No overdraft~% Can't withdraw more than you have: $~a~%"
               (balance)))
           (else
             (remember (- (balance) amount))
             (format #t "New balance: $~a~%" (balance)))))
       ((current-balance self password)
         (if (or (eq? password master-password) (equal? password a-pin))
           (format #t "Your Balance is $~a~%" (balance))
           (bad-password self password)))
       ;; only bank has access to account history
       ((get-account-history self password)
         (if (eq? password master-password)
           history
           (bad-password self password))))))
 
 ;;;------------------------------------------
 ;; bank-account implementation
 ;;-------------------------------------------
 
 (define (make-account a-name an-age a-ssn a-pin initial-balance master-password)
   (object-with-ancestors
     ((customer (make-person a-name an-age a-ssn a-pin))
      (account (make-account-history initial-balance a-pin master-password)))
     ((get-pin self password)
       (if (eq? password master-password)
         a-pin
         (bad-password self password)))
     ((get-account-history self password)
       (operate-as account get-account-history self password))
     ;; our bank is very conservative...
     ((bad-password self bogus-passwd)
       (format #t "~%CALL THE POLICE!!~%"))
     ;; protect the customer as well
     ((ssn self password)
       (operate-as customer ssn self password))
     ((print-obj self port)
       (format port "#<Bank-Customer ~a>~%" (name self)))))
 
 ;;; eof yasos-examples.scm

(define p2 (make-point 1 2))
(define p3 (make-point-3d 4 5 6))
(define fred  (make-person  "Fred"  19 "573-19-4279" 'FadeCafe))
(define sally (make-account "Sally" 26 "629-26-9742" 'FeedBabe 263 'bank-password))

(testeez
 "yasos"
 (test/equal "Size of 2D point"
	     (size-obj p2)
	     2)
 (test/equal "Size of 3D point"
	     (size-obj p3)
	     3)
 (test/equal "A symbol is not an instance"
	     (with-output-to-string (lambda() (print-obj 'mist #t)))
	     "#<NOT-AN-INSTANCE: mist>\n")
 (test/equal "A 2D point is an instance of <point>"
	     (with-output-to-string (lambda() (print-obj p2 #t)))
	     "#<point: 1 2>\n")
 (test/equal "A 2D point is a recognized as point"
	     (point? p2)
	     #t)
 (test/equal "A 2D point is not recognized as a 3D point"
	     (point-3d? p2)
	     #f)
 (test/equal "A 3D point is an instance of <3d-point>"
	     (with-output-to-string (lambda() (print-obj p3 #t)))
	     "#<3d-point: 4 5 6>\n")
 (test/equal "A 3D point is recognized as a point"
	     (point? p3)
	     #t)
 (test/equal "A 3D point is recognized as a 3D point"
	     (point-3d? p3)
	     #t)
 (test/equal "Print an instance of <person>"
	     (with-output-to-string (lambda () (print-obj fred #t)))
	     "#<Person: Fred age: 19>\n")
 (test/equal "Access ssn attribute of <person>"
	     (ssn fred 'FadeCafe)
	     "573-19-4279")
 (test/equal "Instance of <person> recognized as such"
	     (person? fred)
	     #t)
 (test/equal "Instance of <person> not recognized as instance of <bank-account>"
	     (bank-account? fred)
	     #f)
 (test/equal "Print an instance of <bank-account>"
	     (with-output-to-string (lambda () (print-obj sally #t)))
	     "#<Bank-Customer Sally>\n")
 (test/equal "Access ssn attribute of <bank-account>"
	     (ssn sally 'FeedBabe)
	     "629-26-9742")
 (test/equal "Instance of <bank-account> recognized as instance of <person>"
	     (person? sally)
	     #t)
 (test/equal "Instance of <bank-account> recognized as such"
	     (bank-account? sally)
	     #t)
 (test/equal "Current balance"
	     (with-output-to-string (lambda()(current-balance sally 'FeedBabe)))
	     "Your Balance is $263\n")
 (test/equal "Add 200 to balance, then another 300, then withdraw 400"
	     (begin (with-output-to-string (lambda()(add sally 200)))
		    (with-output-to-string (lambda()(add sally 300)))
		    (with-output-to-string (lambda()(withdraw sally 400 'FeedBabe)))
		    (with-output-to-string (lambda()(current-balance sally 'FeedBabe))))
	     "Your Balance is $363\n")
 (test/equal "Account history"
	     (get-account-history sally 'bank-password)
	     '(363 763 463 263) )
 (test/equal "Withdrawal with bank password"
	     (with-output-to-string
	       (lambda () (withdraw sally 150 (get-pin sally 'bank-password))))
	     "New balance: $213\n")
 (test/equal "Account history after withdrawal with bank password"
	     (get-account-history sally 'bank-password)
	     '(213 363 763 463 263) )
 (test/equal "Bogus password"
	     (with-output-to-string (lambda() (ssn fred 'bogus)))
	     "Bad password: bogus\n")
)
